@@ -9,9 +9,8 @@ from scipy.special import gamma, logsumexp, digamma, polygamma, loggamma
 from scipy.optimize import newton
 from .variational_hyperparams import VariationalMixHyperparams as Hyperparams
 from .parameter_bundle import ParameterBundle
-from copy import copy
-from squaredMahaDistance import squaredMahaDistance
-
+from optimized_mstep_functions import squaredMahaDistance
+from .mixture_base_class import MixtureBaseClass 
 
 
 
@@ -86,13 +85,14 @@ from squaredMahaDistance import squaredMahaDistance
 #                   allow algorithm to optimize.
 #converged_     --  Whether the fit converged.
 
-class VariationalStudentMixture():
+class VariationalStudentMixture(MixtureBaseClass):
 
     def __init__(self, n_components = 2, tol=1e-5, max_iter=2000, n_init=1,
             df = 4.0, fixed_df = True, random_state=123, verbose=False,
-            init_type = "k++", scale_inv_prior=None, loc_prior=None,
+            init_type = "kmeans", scale_inv_prior=None, loc_prior=None,
             mean_cov_prior = 1e-2, weight_conc_prior=None, wishart_dof_prior = None,
             max_df = 100):
+        super().__init__()
         self.max_df = max_df
         if self.max_df is None:
             raise ValueError("max_df cannot be None! Set max_df to np.inf if you would like "
@@ -137,103 +137,8 @@ class VariationalStudentMixture():
         self.final_lower_bound = None
         self.resp = None
 
-    #Function to check the user specified model parameters for validity.
-    def check_user_params(self, n_components, tol, reg_covar, max_iter, n_init, df, random_state,
-            init_type):
-        try:
-            n_components = int(n_components)
-            tol = float(tol)
-            n_init = int(n_init)
-            random_state = int(random_state)
-            max_iter = int(max_iter)
-            reg_covar = float(reg_covar)
-            init_type = str(init_type)
-        except:
-            raise ValueError("n_components, tol, max_iter, n_init, reg_covar and random state should be numeric; "
-                    "init_type should be a string.")
-        if df > self.max_df:
-            raise ValueError("Very large values for dof will give results essentially identical to a Gaussian mixture."
-                    "DF = 4 is suggested as a good default. If fixed_df is False, the df will be "
-                             "optimized. Currently you have chosen a starting value for df that is "
-                             "greater than the max_df value you specified.")
-        if df > 1000:
-            raise ValueError("Very large values for dof will give results essentially identical to a Gaussian mixture."
-                    "DF = 4 is suggested as a good default. If fixed_df is False, the df will be "
-                             "optimized. Currently you have chosen a very large value for df > 1000.")
-        if df < 1:
-            raise ValueError("Inappropriate starting value for df!")
-        if max_iter < 1:
-            raise ValueError("Inappropriate value for the maximum number of iterations! Must be >= 1.")
-        if n_init < 1:
-            raise ValueError("Inappropriate value for the number of restarts! Must be >= 1.")
-        if tol <= 0:
-            raise ValueError("Inappropriate value for tol! Must be greater than 0.")
-        if n_components < 1:
-            raise ValueError("Inappropriate value for n_components! Must be >= 1.")
-        if reg_covar < 0:
-            raise ValueError("Reg covar must be >= 0.")
-        if init_type not in ["k++", "kmeans"]:
-            raise ValueError("init_type must be one of either 'k++' or 'kmeans'.")
 
 
-    #Function to check whether the input has the correct dimensionality. This function is
-    #used to check data supplied to self.predict, NOT fitting data, and assumes the model
-    #has already been fitted -- it compares the dimensionality of the input to the dimensionality
-    #of training data. To check training data for validity, self.check_fitting_data is used instead.
-    def check_inputs(self, X):
-        if isinstance(X, np.ndarray) == False:
-            raise ValueError("X must be a numpy array.")
-        #Check first whether model has been fitted.
-        self.check_model()
-        if X.dtype != "float64":
-            raise ValueError("The input array should be of type float64.")
-        if len(X.shape) > 2:
-            raise ValueError("Only 1d or 2d arrays are accepted as input.")
-        x = X
-        if len(x.shape) == 1:
-            x = x.reshape(-1,1)
-        if x.shape[1] != self.location_.shape[1]:
-            raise ValueError("Dimension of data passed does not match "
-                    "dimension of data used to fit the model! The "
-                    "data used to fit the model has D=%s"
-                    %self.location_.shape[1])
-        return x
-
-
-    #Function to check whether the model has been fitted yet. The parameters
-    #are all set at the same time, so we need merely check any one of them (e.g. df_)
-    #to see whether the model has been fitted.
-    def check_model(self):
-        if self.df_ is None:
-            raise ValueError("The model has not been successfully fitted yet.")
-
-
-    #Check data supplied for fitting to make sure it meets basic
-    #criteria. We require that N > 2*D and N > 3*n_components. If the user
-    #supplied a 1d input array, we reshape it to 2d -- this enables us to
-    #cluster 1d input arrays without forcing the user to reshape them -- a
-    #little bit more user-friendly than scikitlearn's classes, which will just raise
-    #a value error if you input a 1d array as X.
-    def check_fitting_data(self, X):
-        if isinstance(X, np.ndarray) == False:
-            raise ValueError("X must be a numpy array.")
-        if X.dtype != "float64":
-            raise ValueError("The input array should be of type float64.")
-        if len(X.shape) > 2:
-            raise ValueError("This class only accepts 1d or 2d arrays as inputs.")
-        if len(X.shape) == 1:
-            x = X.reshape(-1,1)
-        else:
-            x = X
-        if x.shape[0] <= 2*x.shape[1]:
-            raise ValueError("Too few datapoints for dataset "
-            "dimensionality. You should have at least 2 datapoints per "
-            "dimension (preferably more).")
-        if x.shape[0] <= 3*self.n_components:
-            raise ValueError("Too few datapoints for number of components "
-            "in mixture. You should have at least 3 datapoints per mixture "
-            "component (preferably more).")
-        return x
 
 
     #Function for fitting a model using the parameters the user
@@ -309,7 +214,8 @@ class VariationalStudentMixture():
     #param_bundle   --  Object containing all fit parameters and all values needed to calculate
     #                   the lower bound.
     def fitting_restart(self, X, random_state, hyperparams):
-        params = ParameterBundle(X, self.n_components, self.start_df, random_state)
+        params = ParameterBundle(X, self.n_components, self.start_df, random_state,
+                        self.init_type)
         
         #The param_bundle has several expectations that need to be initialized before
         #the first fitting iteration -- this is done by the following function call.
@@ -328,7 +234,7 @@ class VariationalStudentMixture():
             if abs(change) < self.tol:
                 convergence = True
                 break
-            old_lower_bound = copy(new_lower_bound)
+            old_lower_bound = new_lower_bound
             if self.verbose:
                 #print(change)
                 print("Actual lower bound: %s" % new_lower_bound)
@@ -369,6 +275,8 @@ class VariationalStudentMixture():
         sq_maha_dist = np.empty((X.shape[0], self.n_components))
         squaredMahaDistance(X, params.loc_, params.scale_inv_chole_,
                 sq_maha_dist)
+        #sq_maha_dist = self.sq_maha_distance(X, params.loc_, params.scale_inv_chole_)
+        
         assigned_comp = np.argmin(sq_maha_dist, axis=1)
         params.resp[np.arange(X.shape[0]), assigned_comp] = 1.0
         params.a_nm = 0.5 * (params.resp * X.shape[1] + params.df_[np.newaxis,:])
@@ -421,6 +329,8 @@ class VariationalStudentMixture():
     def VariationalEStep(self, X, params, sq_maha_dist):
         squaredMahaDistance(X, params.loc_, params.scale_inv_chole_,
                         sq_maha_dist)
+        #sq_maha_dist = self.sq_maha_distance(X, params.loc_, params.scale_inv_chole_)
+        
         sq_maha_dist = sq_maha_dist * params.wishart_vm[np.newaxis,:] + \
                        X.shape[1] / params.eta_m[np.newaxis,:]
         weighted_loglik = self.variational_loglik(X, params, sq_maha_dist)
@@ -682,52 +592,11 @@ class VariationalStudentMixture():
         return inv_chole_array
 
 
-    #Calculates log p(X | theta) where theta is the current set of parameters but does
-    #not apply mixture weights.
-    #The function returns an array of dim N x K for N datapoints, K mixture components.
-    #It expects to receive the squared mahalanobis distance and the model 
-    #parameters (since during model fitting these are still being updated).
-    def get_loglikelihood(self, X, sq_maha_dist, df_, scale_cholesky_, mix_weights_):
-        sq_maha_dist = 1 + sq_maha_dist / df_[np.newaxis,:]
-        
-        #THe rest of this is just the calculations for log probability of X for the
-        #student's t distributions described by the input parameters broken up
-        #into three convenient chunks that we sum on the last line.
-        sq_maha_dist = -0.5*(df_[np.newaxis,:] + X.shape[1]) * np.log(sq_maha_dist)
-        
-        const_term = loggamma(0.5*(df_ + X.shape[1])) - loggamma(0.5*df_)
-        const_term = const_term - 0.5*X.shape[1]*(np.log(df_) + np.log(np.pi))
-        
-        scale_logdet = [np.sum(np.log(np.diag(scale_cholesky_[:,:,i])))
-                        for i in range(self.n_components)]
-        scale_logdet = np.asarray(scale_logdet)
-        return -scale_logdet[np.newaxis,:] + const_term[np.newaxis,:] + sq_maha_dist
-
-
-
-    '''The remaining functions are called for a fitted model. Each of the functions
-    that the user is expected to call (i.e. the ones described in the documentation)
-    check that the model has been fitted using self.check_model() before performing any
-    calculations; if the model has not been fitted they raise a value error.'''
+    '''The remaining functions are called for a fitted model. They are unique to the 
+    Variational class and hence are not stored under the base class, where functions
+    common to both Variational and EM classes are stored.'''
     
 
-
-
-    #Returns a categorical component assignment for each sample in the input. It calls
-    #predict_proba which performs a model fit check, then assigns each datapoint to
-    #the component with the largest likelihood for that datapoint.
-    def predict(self, X):
-        probs = self.predict_proba(X)
-        return np.argmax(probs, axis=1)
-
-    #Returns the probability that each sample belongs to each component. Used by
-    #self.predict.
-    def predict_proba(self, X):
-        self.check_model()
-        x = self.check_inputs(X)
-        probs = self.get_component_probabilities(x)
-        return probs
-
     #At the end of fitting, with a variational mixture, the user may sometimes find that
     #some of the components were unnecessary and can be killed off. This function
     #enables the user to remove empty clusters, which may be cheaper than re-fitting
@@ -768,128 +637,3 @@ class VariationalStudentMixture():
                             self.scale_inv_cholesky_)
         
 
-    #Returns the average log likelihood (i.e. averaged over all datapoints). Calls
-    #self.score_samples to do the actual calculation. Useful for AIC, BIC. It has
-    #the option to not perform model checks since it is called by AIC and BIC which
-    #perform model checks before calling score.
-    def score(self, X, perform_model_checks = True):
-        return np.mean(self.score_samples(X, perform_model_checks))
-
-    #Returns the per sample log likelihood. Useful if fitting a class conditional classifier
-    #with a mixture for each class.
-    def score_samples(self, X, perform_model_checks = True):
-        if perform_model_checks:
-            self.check_model()
-            X = self.check_inputs(X)
-        return logsumexp(self.get_weighted_loglik(X), axis=1)
-        
-    #Simultaneously fits and makes predictions for the input dataset.
-    def fit_predict(self, X):
-        self.fit(X)
-        return self.predict(X)
-
-    #Gets the locations for the fitted mixture model. self.check_model ensures
-    #model has been fitted and will raise a value error if it hasn't.
-    @property
-    def location(self):
-        self.check_model()
-        return self.location_
-
-    #Setter for the location attribute.
-    @location.setter
-    def location(self, user_assigned_location):
-        self.location_ = user_assigned_location
-
-    #Gets the scale matrices for the fitted mixture model.
-    @property
-    def scale(self):
-        self.check_model()
-        return self.scale_
-
-    #Setter for the scale attribute.
-    @scale.setter
-    def scale(self, user_assigned_scale):
-        self.scale_ = user_assigned_scale
-
-    #Gets the mixture weights for a fitted model.
-    @property
-    def mix_weights(self):
-        self.check_model()
-        return self.mix_weights_
-
-    #Setter for the mix weights.
-    @mix_weights.setter
-    def mix_weights(self, user_assigned_weights):
-        self.mix_weights_ = user_assigned_weights
-
-    #Gets the degrees of freedom for the fitted mixture model.
-    @property
-    def degrees_of_freedom(self):
-        self.check_model()
-        return self.df_
-
-    #Setter for the degrees of freedom.
-    @degrees_of_freedom.setter
-    def degrees_of_freedom(self, user_assigned_df):
-        self.df_ = user_assigned_df
-
-
-    #Returns log p(X | theta) + log mix_weights. This is called by other class
-    #functions which check before calling it that the model has been fitted.
-    def get_weighted_loglik(self, X):
-        sq_maha_dist = np.empty((X.shape[0], self.n_components))
-        squaredMahaDistance(X, self.location_, self.scale_inv_cholesky_,
-                        sq_maha_dist)
-        loglik = self.get_loglikelihood(X, sq_maha_dist, self.df_, self.scale_cholesky_,
-                        self.mix_weights_)
-        return loglik + np.log(self.mix_weights_)[np.newaxis,:]
-
-
-    #Returns the probability that the input data belongs to each component. Used
-    #for making predictions. This is called by other class functions which check before
-    #calling it that the model has been fitted.
-    def get_component_probabilities(self, X):
-        weighted_loglik = self.get_weighted_loglik(X)
-        with np.errstate(under="ignore"):
-            loglik = weighted_loglik - logsumexp(weighted_loglik, axis=1)[:,np.newaxis]
-        return np.exp(loglik)
-
-
-    #Gets the number of parameters (useful for AIC & BIC calculations). Note that df is only
-    #treated as a parameter if df is not fixed. This function is only used by AIC and BIC
-    #which check whether the model has been fitted first so no need to check here.
-    def get_num_parameters(self):
-        num_parameters = self.n_components - 1 + self.n_components * self.location_.shape[1]
-        num_parameters += 0.5 * self.scale_.shape[0] * (self.scale_.shape[1] + 1) * self.scale_.shape[2]
-        if self.fixed_df:
-            return num_parameters
-        else:
-            return num_parameters + self.df_.shape[0]
-
-
-    #Samples from the fitted model with a user-supplied random seed. (It is important not to
-    #use the random seed saved as self.random_state because the user may want to easily update
-    #the random seed when sampling, depending on their needs.)
-    #We sample from the multinomial distribution described by the mixture weights to
-    #determine the number of datapoints per component. Next, we sample from the
-    #chisquare distribution (a chisquare is a special case of a gamma, and remember
-    #that student's t distributions can be described as an infinite scale mixture
-    #of Gaussians). Finally, we sample from a standard normal and shift using
-    #the location and the sample from the chisquare distribution.
-    def sample(self, num_samples = 1, random_seed = 123):
-        if num_samples < 1:
-            raise ValueError("You can't generate less than one sample!")
-        self.check_model()
-        rng = np.random.RandomState(random_seed)
-        samples_per_component = rng.multinomial(n=num_samples, pvals=self.mix_weights_)
-        sample_data = []
-        for i in range(self.n_components):
-            if np.isinf(self.df_[i]):
-                x = 1.0
-            else:
-                x = rng.chisquare(self.df_[i], size=samples_per_component[i])
-            comp_sample = rng.multivariate_normal(np.zeros(self.location_.shape[1]),
-                            self.scale_[:,:,i], size=samples_per_component[i])
-            sample_data.append(self.location_[i,:] + comp_sample / np.sqrt(x)[:,np.newaxis])
-        return np.vstack(sample_data)
-
